/*######     Copyright (c) 1997-2012 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com    ##########################################
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published #
# by the Free Software Foundation; either version 3, or (at your option) any later version. This program is distributed in the hope that #
# it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. #
# See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with this #
# program; If not, see <http://www.gnu.org/licenses/>                                                                                    #
########################################################################################################################################*/
#pragma once

//!!!#include <ia32.h>

namespace Ia32 {

ENUM_CLASS(Ia32Instr) {
	INSTR_MOV,
	INSTR_CALL,
	INSTR_RET,
	INSTR_IRET,
	INSTR_PUSH
} END_ENUM_CLASS(Ia32Instr);


const byte REG_MASK = 0x07,
			REG_GP	= 0x08;


enum Ia32Register {
	REG_A_AX = 0 | REG_GP,
	REG_A_CX = 1 | REG_GP,
};


class Ia32Op {
public:
	enum EType {
		OP_NONE,
		OP_REG,
		OP_OFFSET
	};

	enum EType Type;

	Ia32Register BaseReg;
	Int32 Offset;
	CBool Indirect;

	Ia32Op()
		: Type(OP_NONE)
	{
	}

	Ia32Op(Ia32Register reg)
		:	Type(OP_REG)
		,	BaseReg(reg)
	{
	}

	Ia32Op(Int32 off)
		:	Type(OP_OFFSET)
		,	Offset(off)
	{
	}

	Ia32Op(void *off)
		:	Type(OP_OFFSET)
		,	Offset((Int32)(Int64)off)
	{
	}
};




class Ia32Codegen {
public:
	BinaryWriter Wr;

	Ia32Codegen(class Stream& stm)
		:	Wr(stm)
	{
	}

	void EmitByte(byte b) {
		Wr << b;
	}

	void EmitInt16(Int16 v) {
		Wr << v;
	}

	void EmitInt32(Int32 v) {
		Wr << v;
	}

//!!!R	void Emit(Ia32Instr instr);
	void Emit(Ia32Instr instr, const Ia32Op& op = Ia32Op(), const Ia32Op& op2 = Ia32Op());

	Ia32Op operator[](Ia32Op op) {
		op.Indirect = true;
		return op;
	}
};

} // namespace Ia32














