/*######     Copyright (c) 1997-2012 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com    ##########################################
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published #
# by the Free Software Foundation; either version 3, or (at your option) any later version. This program is distributed in the hope that #
# it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. #
# See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with this #
# program; If not, see <http://www.gnu.org/licenses/>                                                                                    #
########################################################################################################################################*/
#pragma once

namespace Ext {

class CCommandLineInfo { //!!!
	void ParseParamFlag(RCString pszParam);
	void ParseParamNotFlag(RCString pszParam);
	void ParseLast(bool bLast);
public:
	bool m_bShowSplash;
	bool m_bRunEmbedded;
	bool m_bRunAutomated;
	enum { FileNew, FileOpen, FilePrint, FilePrintTo, FileDDE, AppRegister,
		AppUnregister, FileNothing = -1 } m_nShellCommand;

	// not valid for FileNew
	String m_strFileName;

	// valid only for FilePrintTo
	String m_strPrinterName;
	String m_strDriverName;
	String m_strPortName;

	CCommandLineInfo();
	virtual ~CCommandLineInfo() {}
	virtual void ParseParam(RCString pszParam, bool bFlag, bool bLast);
};

#if UCFG_ARGV_UNICODE
typedef wchar_t argv_char_t;
#else
typedef char argv_char_t;
#endif

class CAppBase
#if UCFG_COMPLEX_WINAPP	
	:	public CWinThread //!!! must be Thread
#endif
{
	typedef CAppBase class_type;
public:
	EXT_DATA static bool s_bSigBreak;

	std::vector<String> m_argv;
	int Argc;
	CPointer<argv_char_t *> Argv;
	std::vector<argv_char_t*> m_argvp;

	bool m_bPrintLogo;
	String FileDescription;
	String SVersion;
	String LegalCopyright;
	String Url;
	String m_appName;
#ifdef _WIN32
	ptr<ThreadPool> m_pThreadPool;
#endif

	String GetInternalName();

	String m_appDataDir;

	String get_AppDataDir();
	DEFPROP_GET(String, AppDataDir);


#if UCFG_COMPLEX_WINAPP
	HINSTANCE m_hInstance;

	String m_exeName;
	String m_cmdLine;
	int m_nCmdShow;

	RegistryKey m_keyLM,
		m_keyCU;

	CAppBase();
	void SetCurrentHandles();
	String GetCompanyName();
	String GetRegPath();

	RegistryKey& get_KeyLM() {
		if (!m_keyLM.Valid())
			m_keyLM.Open(HKEY_LOCAL_MACHINE, GetRegPath(), true);
		return m_keyLM;
	}
	DEFPROP_GET(RegistryKey&, KeyLM);

	RegistryKey& get_KeyCU() {
		if (!m_keyCU.Valid())
			m_keyCU.Open(HKEY_CURRENT_USER, GetRegPath(), true);
		return m_keyCU;
	}
	DEFPROP_GET(RegistryKey&, KeyCU);
#else
	static CAppBase *I;

	CAppBase()
		:	m_bPrintLogo(true)
	{
		I = this;
	}

	~CAppBase() {
		I = nullptr;
	}

#endif

#if !UCFG_EXTENDED
	virtual int InitInstance() { return 1; }
	virtual int ExitInstance() {
		return 1;
	}
#endif

	void AFXAPI ParseCommandLine(CCommandLineInfo& rCmdInfo);
protected:
	void ProcessArgv(int argc, argv_char_t *argv[]);
#if !UCFG_ARGV_UNICODE
	void ProcessArgv(int argc, String::Char *argv[]);
#endif

	String m_internalName;
};

AFX_API CAppBase * AFXAPI AfxGetCApp();

class CConApp : public CAppBase, public CCommandLineInfo {
	static CConApp *s_pApp;

	static void __cdecl OnSigInt(int sig);
public:
	CConApp();

	EXT_API int Main(int argc, argv_char_t *argv[]);
	virtual void Usage() {}
	virtual void Execute() {}	
	virtual bool OnSignal(int sig);

	static void AFXAPI SetSignals();
};

#ifdef WIN32

#	if UCFG_ARGV_UNICODE
#		define ext_main wmain
#	else
#		define ext_main main
#	endif
#	define EXT_DEFINE_MAIN(app)	extern "C" int __cdecl ext_main(int argc, argv_char_t *argv[]) { return app.Main(argc, argv); }

#else

#	define EXT_DEFINE_MAIN(app)	extern "C" int __cdecl main(int argc, char *argv[]) {	\
	app.FileDescription = VER_FILEDESCRIPTION_STR;									\
	app.SVersion			= VER_PRODUCTVERSION_STR;									\
	app.LegalCopyright = VER_LEGALCOPYRIGHT_STR;										\
	app.Url = VER_EXT_URL;															\
	return app.Main(argc, argv); }

#endif

#ifdef WIN32

class CDocTemplate;
class CDocManager;

class EXTAPI CWinApp : public CAppBase {
#if UCFG_EXTENDED
	DECLARE_MESSAGE_MAP()
#endif
public:
	CWinApp(RCString lpszAppName = "");
	virtual ~CWinApp();
	virtual int InitInstance();
protected:
	int m_nWaitCursorCount;
	HCURSOR m_hcurWaitCursorRestore;

#if UCFG_COMPLEX_WINAPP

protected:
#if UCFG_EXTENDED && !UCFG_WCE	

	AutoPtr<CRecentFileList> m_pRecentFileList;

	void LoadStdProfileSettings(UINT nMaxMRU = _AFX_MRU_COUNT); // load MRU file list and last preview state
	void SaveStdProfileSettings();  // save options to .INI file
#endif
	void SetRegistryKey(RCString registryKey);
	afx_msg void OnFileNew();
	afx_msg void OnFileOpen();
	afx_msg void OnHelp();
	afx_msg void OnAppExit();
public:
	String m_registryKey;
	String m_profileName;
	AutoPtr<Object> m_pDocManager;
	AutoPtr<CCommandLineInfo> m_pCmdInfo;


	bool ProcessShellCommand(CCommandLineInfo& rCmdInfo);
#if UCFG_WND
	virtual int DoMessageBox(RCString lpszPrompt, UINT nType, UINT nIDPrompt);
#endif
#if UCFG_GUI
	void AddDocTemplate(CDocTemplate* pTemplate);
	virtual void DoWaitCursor(int nCode);
	EXT_API HICON LoadIcon(const CResID& resID) const;
#if !UCFG_WCE
	virtual void AddToRecentFileList(RCString pathName);  // add to MRU
#endif
#if UCFG_WIN32_FULL
	virtual CDocument* OpenDocumentFile(RCString lpszFileName);
#endif
	virtual bool Register();
	virtual bool Unregister();
#endif
	void RegisterShellFileTypes(BOOL bCompat);
	void UnregisterShellFileTypes();
	virtual void OnAbort() {}
	BOOL SaveAllModified();
	void HideApplication();
	void CloseAllDocuments(BOOL bEndSession);
	bool DoPromptFileName(String& fileName, UINT nIDSTitle, DWORD lFlags, bool bOpenFileDialog, CDocTemplate* pTemplate);

	CDocManager *get_DocManager();
	DEFPROP_GET(CDocManager*, DocManager);

	LRESULT ProcessWndProcException(RCExc e, const MSG *pMsg);
	String GetProfileString(LPCTSTR lpszSection, LPCTSTR lpszEntry, LPCTSTR lpszDefault = 0);
	void WriteProfileString(LPCTSTR lpszSection, LPCTSTR lpszEntry, LPCTSTR lpszValue);
	void Enable3dControls();
	void GetAppRegistryKey(RegistryKey& key);

#endif
};



const int UWM_NOTIFYICON = WM_USER+1;
const int UWM_INVALID_REGCODE = WM_USER+10;


//!!!FILETIME AFXAPI StringToFileTime(RCString s);

//!!!#	ifndef _EXT 

class EXT_CLASS CUpgradeBase : public CSeparateThread {
	virtual void BreakProgram(bool bClearRegCode = true) {}
public:
	static CUpgradeBase *I;
	static bool s_bUseStdioRedirector;

	CBool m_bRegistered,
		m_bEnabled, m_bRegcodeFound;
	LONG m_bCreated;
	Version m_myVer;
	CUpgradeBase *m_prev;
	CBool AutoGetwork;

	CUpgradeBase()
		:	m_bEnabled(true)
		,	m_bCreated(false)
		,	m_prev(I)
	{
		I = this;
	}

	~CUpgradeBase() {
		I = m_prev;
	}

	void Notify(RCString s);
	void NotifyRelURL(RCString relUrl) { NotifyURL(Http()+relUrl); }
	virtual void Create() {
	}
	virtual bool Check(int options = 5, CWnd *pWnd = 0);
	virtual int GetLicenses() { return 0;}
	virtual bool IsRegistered() { return m_bEnabled; }
	virtual void NotifyURL(RCString url) {}
	virtual String Http() { return MANUFACTURER_HTTP_URL "/"; }
	virtual String News() { return MANUFACTURER_NEWS_URL "/"; }
	virtual int CheckNewVersion(int n = 2, bool bForceAsk = false) { return IDCANCEL; }

	virtual String GetRandomSiteHost() { return MANUFACTURER_DOMAIN; }  
	String GetRandomBaseHttp() { return "http://"+GetRandomSiteHost()+"/"; }

	virtual String Email() { return MANUFACTURER_EMAIL; }
	virtual String Fax() { return MANUFACTURER_FAX; }
	virtual String GetNotifyURL() { return MANUFACTURER_NOTIFY_URL; }
	virtual void OnInvalidRegcode(int wp) { Check(7); }
	virtual void OnRegister(CWnd *pWnd = 0) {}
	virtual void Stop() {}
	virtual bool EnabledNotifications() { return true; }
	virtual void StartWorking() {}
	virtual void StopWorking() {}
	virtual bool IsComputationallyPowerful() { return false; }
	virtual int GetFreeUsing(CWinApp& app) { return 0; }
};



inline bool AFXAPI IsRegistered() { return CUpgradeBase::I->IsRegistered(); }


class CConsoleStreamHook {
	FILE *m_file;
public:
	int m_descPrev;
	int m_arDesc[2];

	CConsoleStreamHook(FILE *file = stdout, size_t size = 0);
	virtual ~CConsoleStreamHook();
	void CreateFromFile(FILE *file, size_t size = 0);
	void CloseInput();
	Blob ReadToEnd();
};

class CStreamHookThread : public Thread, public CConsoleStreamHook {
	void Stop() {
		CloseInput();
		Thread::Stop();
	}

	void Execute();
public:
	CStreamHookThread(CThreadRef *tr, FILE *file)
		:	Thread(tr)
		,	CConsoleStreamHook(file)
	{}
};

class CConsoleStreamsCodepageConverter {
	CThreadRef m_tr;
public:
	CConsoleStreamsCodepageConverter();
	~CConsoleStreamsCodepageConverter();
};

#if !UCFG_WCE && UCFG_EXTENDED

class CNewConsoleOut : public std::ostream {
	Process m_proc;

	class MyStreambuf : public std::streambuf {
		CNewConsoleOut *m_out;

		int overflow(int c) {
			if (c != EOF) {
				m_out->m_proc.StandardInput.WriteBuffer(&c, 1);
				return true;
			}
			return std::streambuf::overflow(c);
		}
	public:
		MyStreambuf(CNewConsoleOut *out)
			:	m_out(out)
		{}
	};

public:
	CNewConsoleOut()
		:	std::ostream(&m_sbuf)
		,	m_sbuf(this)
	{
		ProcessStartInfo psi;
		psi.FileName = "cat";
		if (!AllocConsole())
			psi.Flags = CREATE_NEW_CONSOLE;
		psi.RedirectStandardInput = true;
		m_proc.StartInfo = psi;
		m_proc.Start();
	}
private:
	MyStreambuf m_sbuf;
};

#endif

class CDllServer {
public:
	static CDllServer *I;

	CDllServer() {
		I = this;
	}

	virtual ~CDllServer() {
		I = 0;
	}

	virtual HRESULT OnRegister() {
#if UCFG_COM_IMPLOBJ
		return COleObjectFactory::UpdateRegistryAll();
#else
		return S_OK;
#endif
	}

	virtual HRESULT OnUnregister() {
#if UCFG_COM_IMPLOBJ
		return COleObjectFactory::UpdateRegistryAll(false);
#else
		return S_OK;
#endif
	}
};



//!!!#	endif // _EXT

#endif // WIN32

EXT_API String AFXAPI RtfToText(RCString rtf);
EXT_API String AFXAPI HtmlToText(RCString html);


} // Ext::
